/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include "is_types.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)is_bal.c	1.4	10:34:10	2/26/92";
#endif

extern	int	db_errno;

int
isam_balance (file, index, block, node, pack)
file_t	*file;
int	index;
off_t	block;
keyblk_t *node;
int	pack;
{
	keyblk_t left, center, right;
	keyblk_t *keyblk;
	sort_key_t *key = &file->f_index->i_keys[index];
	off_t	parent;
	off_t	sibbling;
	int	size = key->k_key_size + sizeof block;
	int	max = (file->f_bsize - (3 * sizeof (long) + sizeof (int))) /
			size;
	int	min = max / 3;
	int	do_left = 1,
		do_right = 1;
	char	*cp;
	int	offset;
	int	i;

	/*
	 * A tree can only be balanced if the file is locked.
	 */
	
	if (file->f_locked != F_WRLCK)
		return 0;

	/*
	 * Get the block if they didn't pass it to us
	 */

	if (node == 0) {
		keyblk = &center;
		if (isam_read (file, block, keyblk))
			return -1;
	} else
		keyblk = node;

	/*
	 * See if it is possible to pack this node any more.  If
	 * the caller wants to try, ignore the 1/3 node threshhold.
	 * Otherwise, if there are no left or right sibblings, the
	 * node can't be filled since there is no place to get the
	 * keys from ...
	 */

	if ((! pack && keyblk->kb_cnt > min) ||
			(! keyblk->kb_right && ! keyblk->kb_left))
		return 0;

	/*
	 * Get any right sibbling, or reset the flag saying there
	 * is nothing over to the left.
	 */

	if (keyblk->kb_left) {
		if (isam_read (file, keyblk->kb_left, &left))
			return -1;
	} else
		do_left = 0;

	/*
	 * Get any left sibbling, or reset the flag saying there
	 * is nothing over to the right.
	 */

	if (keyblk->kb_right) {
		if (isam_read (file, keyblk->kb_right, &right))
			return -1;
	} else
		do_right = 0;

	/*
	 * Set the upper limit.  We don't want to fill the node
	 * too full, unless requested to, since that will cause
	 * a split the first time the node is added to.  So leave
	 * some room for additions.
	 */

	if (! pack)
		max = 2 * min;

	/*
	 * Pick which side to do by finding which combination of
	 * keys is going to fit best.  Pick more empty nodes over
	 * more full ones since this improves the packed-ness more.
	 */

	if (do_right && do_left) {
		if (right.kb_cnt < left.kb_cnt &&
				right.kb_cnt + keyblk->kb_cnt < max)
			do_left = 0;
		else if (left.kb_cnt + keyblk->kb_cnt < max)
			do_right = 0;
		else
			return 0;
	}
	if (do_left) {

		/*
		 * Combine the keys in the node to the left with
		 * the keys in the current block.  Both sets of
		 * keys must fit in the new block.
		 */

		if (keyblk->kb_cnt + left.kb_cnt > max)
			return 0;

		cp = left.kb_data + left.kb_cnt * size;
		memcpy (cp, keyblk->kb_data, keyblk->kb_cnt * size);
		keyblk->kb_cnt += left.kb_cnt;
		memcpy (keyblk->kb_data, left.kb_data, keyblk->kb_cnt * size);

		/*
		 * Fix the index information so my position isn't
		 * lost
		 */

		if (file->f_key_index == index) {
			if (file->f_key_block == block)
				file->f_key_slot += left.kb_cnt;
			else if (file->f_key_block = keyblk->kb_left)
				file->f_key_block = block;
		}

		/*
		 * The contents of the left block will be written
		 * back where the center block is located.  The
		 * left sibbling of the left block has to be
		 * updated to point to its new right sibbling.
		 */

		if (left.kb_left) {
			if (isam_read (file, left.kb_left, &right))
				return -1;

			right.kb_right = block;
			if (isam_write (file, left.kb_left, &right))
				return -1;
		}

		/*
		 * The parent of the left node needs to have the
		 * entry for the left node deleted.
		 */

		if (left.kb_prev) {
			if (__isam_delete_block (file, key,
					left.kb_prev, keyblk->kb_left))
				return -1;

			left.kb_prev = keyblk->kb_prev;
		}

		/*
		 * Free the old left node
		 */

		if (isam_free (file, keyblk->kb_left))
			return -1;

		keyblk->kb_left = left.kb_left;

		/*
		 * Write out the new center node
		 */

		if (isam_write (file, block, keyblk))
			return -1;

		/*
		 * The parent of the center node needs to have the
		 * key value updated.
		 */

		if (keyblk->kb_prev) {
			if (__isam_fix (file, key, keyblk, block,
					&right) < 0)
				return -1;
		}
		parent = keyblk->kb_prev;
	} else if (do_right) {

		/*
		 * Combine the keys in the node to the right with
		 * the keys in the current block.  Both sets of
		 * keys must fit in the new block.
		 */

		if (keyblk->kb_cnt + right.kb_cnt > max)
			return 0;

		cp = keyblk->kb_data + keyblk->kb_cnt * size;
		memcpy (cp, right.kb_data, right.kb_cnt * size);
		keyblk->kb_cnt += right.kb_cnt;

		/*
		 * Fix the index information so my position isn't
		 * lost
		 */

		if (file->f_key_index == index) {
			if (file->f_key_block == keyblk->kb_right) {
				file->f_key_block = block;
				file->f_key_slot += (keyblk->kb_cnt -
							right.kb_cnt);
			}
		}

		/*
		 * Free the old right node
		 */

		if (isam_free (file, keyblk->kb_right))
			return -1;

		/*
		 * The contents of the right block will be written
		 * back where the center block is located.  The
		 * right sibbling of the right block has to be
		 * updated to point to its new left sibbling.
		 */

		if (right.kb_right) {
			if (isam_read (file, right.kb_right, &left))
				return -1;

			left.kb_left = block;
			if (isam_write (file, right.kb_right, &left))
				return -1;
		}
		sibbling = keyblk->kb_right;
		keyblk->kb_right = right.kb_right;

		/*
		 * Write out the updated center node
		 */

		if (isam_write (file, block, keyblk))
			return -1;

		/*
		 * The parent of the right node needs to have the
		 * entry for the right node deleted.  No other change
		 * needs to be made because the parent still points
		 * to the correct remaining child.
		 */

		if (right.kb_prev) {
			if (__isam_delete_block (file, key,
				right.kb_prev, sibbling))
				return -1;
		}
		parent = keyblk->kb_prev;
	}
	if (parent) {
		if (isam_read (file, parent, &right))
			return -1;

		return isam_balance (file, index, parent, &right, pack);
	}
	return 0;
}
