/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include "is_types.h"
#include "isam.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)is_compare.c	1.6	11:48:45	1/29/92";
#endif

field_t	dummy = {
	T_LONG,	sizeof (off_t), 0,	".recno"
};

extern	int	db_errno;

int
isam_key_compare (schema, key, key_1, key_2, unique)
schema_t *schema;
sort_key_t *key;
VOIDP	*key_1;
VOIDP	*key_2;
int	unique;
{
	short	s1, s2;
	int	i1, i2;
	long	l1, l2;
	float	f1, f2;
	double	d1, d2;
	char	*cp1, *cp2;
	char	*left = (char *) key_1,
		*right = (char *) key_2;
	int	cnt;
	int	rc = 0;
	int	max;
	int	i;
	int	order = 0;
	field_t	*field;

	if (key->k_unique || unique == IS_UNIQUE)
		max = key->k_field_count;
	else
		max = key->k_field_count - 1;

	assert (max >= 0);

	for (cnt = 0;cnt < max;cnt++) {
		rc = 0;
		if (! key->k_unique && cnt == (key->k_field_count - 1)) {
			field = &dummy;
			order = 0;
		} else {
			i = key->k_fields[cnt] & ~IS_DESCENDING;
			order = key->k_fields[cnt] & IS_DESCENDING;
			field = &schema->s_fields[i];
		}
		switch (field->f_type) {
			case T_CHAR:
				rc = strncmp (left, right, field->f_size);
				if (rc != 0)
					goto done;
				left += field->f_size;
				right += field->f_size;
				break;
			case T_SHORT:
				memcpy (&s1, left, sizeof s1);
				memcpy (&s2, right, sizeof s2);
				if (s1 != s2) {
					rc = s1 < s2 ? -1:1;
					goto done;
				}
				left += sizeof s1;
				right += sizeof s2;
				break;
			case T_INT:
				memcpy (&i1, left, sizeof i1);
				memcpy (&i2, right, sizeof i2);
				if (i1 != i2) {
					rc = i1 < i2 ? -1:1;
					goto done;
				}
				left += sizeof i1;
				right += sizeof i2;
				break;
			case T_LONG:
			case T_DATE:
			case T_SERIAL:
				memcpy (&l1, left, sizeof l1);
				memcpy (&l2, right, sizeof l2);
				if (l1 != l2) {
					rc = l1 < l2 ? -1:1;
					goto done;
				}
				left += sizeof l1;
				right += sizeof l2;
				break;
			case T_FLOAT:
				memcpy (&f1, left, sizeof f1);
				memcpy (&f2, right, sizeof f2);
				if (f1 != f2) {
					rc = f1 < f2 ? -1:1;
					goto done;
				}
				left += sizeof f1;
				right += sizeof f2;
				break;
			case T_DOUBLE:
			case T_MONEY:
				memcpy (&d1, left, sizeof d1);
				memcpy (&d2, right, sizeof d2);
				if (d1 != d2) {
					rc = d1 < d2 ? -1:1;
					goto done;
				}
				left += sizeof d1;
				right += sizeof d2;
				break;
			default:
#ifndef	NDEBUG
				abort (field->f_type != T_CHAR);
#endif
				db_errno = DB_BAD_DB;
				break;
		}
	}
done:
	if (order == IS_DESCENDING)
		rc = - rc;

	return rc;
}
