/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include "is_types.h"

#ifdef BSD
#define memcpy(a, b, c) bcopy((b), (a), (c))
#endif /* BSD */

#ifndef lint
static	char	sccsid[] = "@(#)is_lookup.c	1.5	10:33:01	2/26/92";
#endif

extern	int	db_errno;

long
isam_previous (file, index)
file_t	*file;
int	index;
{
	union {
		keyblk_t u_keyblk;
		char	u_buf[MAXBLOCKSIZE];
	} buf;
	char	*keys = buf.u_keyblk.kb_data;
	keyblk_t *keyblk = &buf.u_keyblk;
	long	block;
	long	ticket;
	int	size;
	int	was_locked;

	/*
	 * Default to the current index if index == -1
	 */

	if (index == -1)
		index = file->f_current_index;

	/*
	 * Validate the parameters ...
	 */

	if (file == 0 || index == 0 || index != file->f_key_index ||
			file->f_current_record < 0)
		return (db_errno = DB_INVALID), -1;

	size = file->f_index->i_keys[index].k_key_size + sizeof block;

	/*
	 * Make sure the index didn't change
	 *
	 *	NEEDSWORK - should re-read the current record and then
	 *	look it up.  That will refresh the key pointers.
	 */

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;
	}
	if (! was_locked && ticket != file->f_ticket)
		goto fail;

	/*
	 * Read the last key block back in.  If f_key_block and
	 * f_key_slot are both zero, assume the index has been
	 * rewound and EOF is being asked for.
	 */

	if (file->f_key_block == 0 && file->f_key_slot == 0) {
		for (block = index;block > 0;) {
			file->f_key_block = block;
			if (isam_read (file, block, keyblk))
				goto fail;

			file->f_key_slot = keyblk->kb_cnt - 1;
			memcpy (&block, keys + file->f_key_slot * size,
				sizeof block);
		}
		goto found;
	}
	block = file->f_key_block;

	if (isam_read (file, block, keyblk))
		goto fail;

	/*
	 * Find out if the last key was the first key or not
	 */

	if (file->f_key_slot == 0) {
		if (keyblk->kb_left == 0) {
			db_errno = DB_BOF;
			goto fail;
		}
		block = file->f_key_block = keyblk->kb_left;
		if (isam_read (file, block, keyblk))
			goto fail;

		file->f_key_slot = keyblk->kb_cnt - 1;
	} else
		file->f_key_slot--;

	/*
	 * Get the record number out of that block
	 */

	memcpy (&block, keys + file->f_key_slot * size, sizeof block);

found:
	if (file->f_key_buffer)
		memcpy (file->f_key_buffer, keyblk, file->f_bsize);

	if (! was_locked) {
		isam_end_ticket (file);
		file->f_ticket = ticket;
	}
	return file->f_current_record = - block;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

long
isam_next (file, index)
file_t	*file;
int	index;
{
	union {
		keyblk_t u_keyblk;
		char	u_buf[MAXBLOCKSIZE];
	} buf;
	char	*keys = buf.u_keyblk.kb_data;
	keyblk_t *keyblk = &buf.u_keyblk;
	long	block;
	long	ticket;
	int	size;
	int	was_locked;

	/*
	 * Default to the current index if index == -1
	 */

	if (index == -1)
		index = file->f_current_index;

	/*
	 * Validate the parameters ...
	 */

	if (file == 0 || index == 0 || index != file->f_key_index ||
			file->f_current_record < 0)
		return (db_errno = DB_INVALID), -1;

	/*
	 * Make sure the index didn't change
	 *
	 *	NEEDSWORK - this should re-read the current record
	 *	and look it up again.  this will refresh the key
	 *	pointers.
	 */

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;
	}
	if (! was_locked && ticket != file->f_ticket)
		goto fail;

	/*
	 * Read the last key block back in
	 */

	if ((block = file->f_key_block) == 0) {
		block = index;
		while (block > 0) {
			file->f_key_block = block;
			if (isam_read (file, block, keyblk))
				goto fail;

			memcpy (&block, keyblk->kb_data, sizeof block);
		}
		block = file->f_key_block;
		file->f_key_slot = -1;
	} else if (isam_read (file, block, keyblk))
		goto fail;

	/*
	 * Find out if the last key was the last key in the block or not
	 */

	if (file->f_key_slot == keyblk->kb_cnt - 1) {
		if (keyblk->kb_right == 0) {
			db_errno = DB_EOF;
			goto fail;
		}
		block = file->f_key_block = keyblk->kb_right;
		if (isam_read (file, block, keyblk))
			goto fail;

		file->f_key_slot = 0;
	} else
		file->f_key_slot++;

	/*
	 * Get the record number out of that block
	 */

	size = file->f_index->i_keys[index].k_key_size + sizeof block;
	memcpy (&block, keys + file->f_key_slot * size, sizeof block);
	if (file->f_key_buffer)
		memcpy (file->f_key_buffer, keyblk, file->f_bsize);

	if (! was_locked) {
		file->f_ticket = ticket;
		isam_end_ticket (file);
	}
	return file->f_current_record = - block;

fail:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

long
__isam_lookup (file, key, root, data, match, unique)
file_t	*file;
sort_key_t *key;
long	root;
VOIDP	*data;
int	match;
int	unique;
{
	keyblk_t buf;
	char	*keys = buf.kb_data;
	keyblk_t *keyblk = &buf;
	long	block,
		last_block;
	long	ticket;
	int	size = key->k_key_size + sizeof block;
	int	cnt;
	int	i;
	int	rc;
	int	was_locked;

	if (! (was_locked = file->f_locked)) {
		if ((ticket = isam_get_ticket (file)) == -1)
			return -1;
	}

new_root:

	/*
	 * The desired record is somewheres in the tree rooted by
	 * "root".  Scan the entire block to find either the exact
	 * record, or the root of the next tree to try.
	 */

	if (isam_read (file, root, keyblk))
		goto not_found;

	if ((cnt = keyblk->kb_cnt) == 0) {
		db_errno = DB_NO_MATCH;
		goto not_found;
	}
	for (block = i = 0;i < cnt;i++) {

		/*
		 * Scan to the right until you find a key that is
		 * greater than or equal to the desired record.
		 */

		rc = isam_key_compare (file->f_schema, key, data,
				keys + i * size + sizeof block, unique);
		file->f_key_match = rc;
		last_block = block;
		memcpy (&block, keys + i * size, sizeof block);

		/*
		 * Nope, this key is too small - the desired key
		 * was bigger than it, so there must be a key
		 * further right that is greater than the current
		 * key and still possibly less than the desired key.
		 */

		if (rc > 0)
			continue;

		if (rc == 0) {
			if (match == IS_GREATER) {
				continue;
			} else if (match == IS_LESS) {
				if (i > 0) {
					if (last_block > 0) {
						root = last_block;
						goto new_root;
					} else {
						file->f_key_block = root;
						file->f_key_slot = i - 1;
						file->f_current_record =
							- last_block;

						goto found;
					}
				} else {
					if (keyblk->kb_left > 0) {
						root = keyblk->kb_left;
						goto new_root;
					}
					db_errno = DB_NO_MATCH;
					goto not_found;
				}
			} else {
				if (block < 0) {
					file->f_key_block = root;
					file->f_key_slot = i;
					file->f_current_record = - block;
					goto found;
				}
				if (i > 0) {
					root = last_block;
					goto new_root;
				} else {
					if (keyblk->kb_left > 0)
						root = keyblk->kb_left;
					else
						root = block;

					goto new_root;
				}
			}
		} else if (rc < 0) {
			if (match == IS_GREATER || match == IS_GREATEREQ) {
				if (block > 0) {
					if (i > 0) {
						root = last_block;
					} else {
						root = block;
					}
					goto new_root;
				} else {
					file->f_key_block = root;
					file->f_key_slot = i;
					file->f_current_record = - block;
					goto found;
				}
			} else if (match == IS_LESS || match == IS_LESSEQ) {
				if (i > 0) {
					if (last_block > 0) {
						root = last_block;
						goto new_root;
					} else {
						file->f_key_block = root;
						file->f_key_slot = i - 1;
						file->f_current_record =
							- last_block;
						goto found;
					}
				} else {
					if (keyblk->kb_left > 0) {
						root = keyblk->kb_left;
						goto new_root;
					}
				}
			} else {
				if (last_block > 0) {
					root = last_block;
					goto new_root;
				} else {
					db_errno = DB_NO_MATCH;
					goto not_found;
				}
			}
		}
	}

	/*
	 * You can always go down, which is kind of like going to the
	 * right in that every key below the current key is greater
	 * than or equal to the current key.  If we can't go down, we
	 * move to the right on the bottom most level in the index.
	 */

	if (block > 0) {
		root = block;
		goto new_root;
	} else if (keyblk->kb_right > 0) {
		root = keyblk->kb_right;
		goto new_root;
	}

	/*
	 * If I was searching for the largest record, I would have used
	 * IS_LESS or IS_LESSEQ and wound up all the way in the right
	 * most lower most spot.  This is that spot ...
	 */

	if (match == IS_LESS || match == IS_LESSEQ) {
		file->f_key_block = root;
		if (i == cnt)
			i--;
		file->f_key_slot = i;
		file->f_current_record = - block;
		goto found;
	}
	db_errno = DB_NO_MATCH;
	goto not_found;

found:
	if (file->f_key_buffer)
		memcpy (file->f_key_buffer, &buf, file->f_bsize);

	if (! was_locked) {
		isam_end_ticket (file);
		file->f_ticket = ticket;
	}
	return file->f_current_record;

not_found:
	if (! was_locked)
		isam_end_ticket (file);

	return -1;
}

long
isam_lookup (file, index, data, match, unique)
file_t	*file;
int	index;
VOIDP	*data;
int	match;
int	unique;
{
	int	i;
	sort_key_t *key;

	/*
	 * Default the index if asked to
	 */

	if (index == -1)
		index = file->f_current_index;

	/*
	 * Find the index to be used.  It must exist, or else
	 * an error is returned.  Block zero (and thus index
	 * zero) is reserved for use as a free index block chain.
	 * Validate match while in the neighborhood.
	 */

	if (index <= 0 || index > file->f_index->i_key_count)
		return (db_errno = DB_INVALID), -1;

	if (match != IS_EQUAL && match != IS_LESS && match != IS_LESSEQ &&
			match != IS_GREATER && match != IS_GREATEREQ)
		return (db_errno = DB_INVALID), -1;

	key = &file->f_index->i_keys[index];

	/*
	 * The base of each index begins at the block numbered
	 * the same as the index number.  Thus, index 1 begins
	 * in block 1.  Index 7 begins in block 7.  There is no
	 * index 0 because it holds the free block chain.
	 */

	return __isam_lookup (file, key, (long) index, data, match, unique);
}
